/*****************************************************************************/
/*
 * Linux driver for Easylon USB Interfaces
 *
 * Copyright (c)2003 Gesytec GmbH, Volker Schober (info@gesytec.de)
 * Licensend for use with Easylon USB Interfaces by Gesytec
 *
 * Please compile lonusb.c
 *
 */
 /*****************************************************************************/

static int lonusb_dl_formware(struct usb_device *usbdev)
{
	const struct firmware *fw;
	const char *fname = FIRMWARE_FX2;
	int ret;
/*
	switch (le16_to_cpu(usbdev->descriptor.idProduct))
    {
	case GESYTEC_LPM2_PID:
		fname = FIRMWARE_FX2;
		break;
	default:
		dev_err(&usbdev->dev, "Missing FW for id: %u\n", le16_to_cpu(usbdev->descriptor.idProduct));
		return;
	};
*/
	ret = request_firmware(&fw, fname, &usbdev->dev);
	if (ret)
    {
		dev_err(&usbdev->dev, "Can't load firmware %s, error: %d\n", fname, ret);
		return ret;
	}

	ret = ezusb_load_firmware(usbdev, fw->data);
	release_firmware(fw);
	if (ret)
		dev_err(&usbdev->dev, "Firmware download failed %d\n", ret);
	return ret;
}

static int CreateIntUrb(PDEVICE_EXTENSION dev)
    {
    int pipe;
#if 0
    int ret = usb_reset_configuration(dev->usbdev);
    if (ret < 0)
        {
        err("usb_reset_configuration() failed");
        return ret;
        }
#endif

    dev->read_urb = usb_alloc_urb(0, GFP_KERNEL);
    if (!dev->read_urb)
        {
        err("No free (read) urbs available");
        return -ENOMEM;
        }
    pipe = usb_rcvintpipe(dev->usbdev, dev->intr_in_ep_addr);
    // dev->bulk_in_size = endpoint->wMaxPacketSize;
    dev->bulk_in_size = usb_maxpacket(dev->usbdev, pipe, usb_pipeout(pipe));
    dbg("intpipe=%X buffer_size=%d", pipe, dev->bulk_in_size);

    dev->intr_in_buff = kmalloc(dev->bulk_in_size, GFP_KERNEL);
    if (!dev->intr_in_buff)
        {
        err("Couldn't allocate intr_in_buff");
        return -ENOMEM;
        }
    usb_fill_int_urb(dev->read_urb, dev->usbdev, pipe,
        dev->intr_in_buff, dev->bulk_in_size,
        lvu_read_int_callback, dev,
        dev->bInterval);

    return 0;
    }


static BOOLEAN reset_neuron(PDEVICE_EXTENSION pdx, BOOLEAN resmes, BOOLEAN doreset, BYTE reset_cause)
    {
    BYTE buffer[8];
    ULONG ret;
    ULONG TimeStamp = 0;
    ULONG Stat = 0;
    ULONG LonStat = 0;
    ULONG ByteCounter = 0;
    ULONG MsgCounter = 0;

    memset(buffer, 0x00, sizeof(buffer));

    pdx->reset_cause = reset_cause;

    TimeStamp     = READ_IO_ULONG(CURR_RECV_TIME_PORT);
    Stat          = READ_IO_ULONG(STAT_PORT);
    LonStat       = READ_IO_ULONG(LON_STAT_PORT);
    ByteCounter   = READ_IO_ULONG(BYTE_COUNTER_PORT);
    MsgCounter    = READ_IO_ULONG(MSG_COUNTER_PORT);

    dbg("reset_neuron(1): TimeStamp=%08X LonStat=%08X Stat=%08X ByteCounter=%02X  MsgCounter=%02X",
        TimeStamp, LonStat, Stat, ByteCounter, MsgCounter);

    ret = Ezusb_VendorRequest (pdx->usbdev, VR_RESET_NEURON, 0x0000, 0x0000, sizeof(buffer), 0, buffer);

    dbg("Ezusb_VendorRequest(VR_RESET_NEURON)=%X buffer=%02X %02X %02X %02X %02X %02X %02X %02X", ((unsigned int)(ret)),
        ((unsigned int)(buffer[0])), ((unsigned int)(buffer[1])),
        ((unsigned int)(buffer[2])), ((unsigned int)(buffer[3])),
        ((unsigned int)(buffer[4])), ((unsigned int)(buffer[5])),
        ((unsigned int)(buffer[6])), ((unsigned int)(buffer[7])));

    TimeStamp     = READ_IO_ULONG(CURR_RECV_TIME_PORT);
    Stat          = READ_IO_ULONG(STAT_PORT);
    LonStat       = READ_IO_ULONG(LON_STAT_PORT);
    ByteCounter   = READ_IO_ULONG(BYTE_COUNTER_PORT);
    MsgCounter    = READ_IO_ULONG(MSG_COUNTER_PORT);

    dbg("reset_neuron(): TimeStamp=%08X LonStat=%08X Stat=%08X ByteCounter=%02X  MsgCounter=%02X",
        ((unsigned int)(TimeStamp)), ((unsigned int)(LonStat)), ((unsigned int)(Stat)),
        ((unsigned int)(ByteCounter)), ((unsigned int)(MsgCounter)));

    // initialize neuron chip
    WRITE_IO_ULONG(CTRL_PORT, NLON_RESET | HOST_RESET);

//    pdx->LastMsgCounter = 0;
//    pdx->LastMsgTime    = KeQueryInterruptTime();

    TimeStamp     = READ_IO_ULONG(CURR_RECV_TIME_PORT);
    Stat          = READ_IO_ULONG(STAT_PORT);
    LonStat       = READ_IO_ULONG(LON_STAT_PORT);
    ByteCounter   = READ_IO_ULONG(BYTE_COUNTER_PORT);
    MsgCounter    = READ_IO_ULONG(MSG_COUNTER_PORT);

    dbg("reset_neuron(): TimeStamp=%08X LonStat=%08X Stat=%08X ByteCounter=%02X  MsgCounter=%02X",
        ((unsigned int)(TimeStamp)), ((unsigned int)(LonStat)), ((unsigned int)(Stat)),
        ((unsigned int)(ByteCounter)), ((unsigned int)(MsgCounter)));

//     KeStallExecutionProcessor(1);

    WRITE_IO_ULONG(CTRL_PORT, NLON_RESET | HOST_RESET);

    TimeStamp     = READ_IO_ULONG(CURR_RECV_TIME_PORT);
    Stat          = READ_IO_ULONG(STAT_PORT);
    LonStat       = READ_IO_ULONG(LON_STAT_PORT);
    ByteCounter   = READ_IO_ULONG(BYTE_COUNTER_PORT);
    MsgCounter    = READ_IO_ULONG(MSG_COUNTER_PORT);

    dbg("reset_neuron(): TimeStamp=%08X LonStat=%08X Stat=%08X ByteCounter=%02X  MsgCounter=%02X",
        ((unsigned int)(TimeStamp)), ((unsigned int)(LonStat)), ((unsigned int)(Stat)),
        ((unsigned int)(ByteCounter)), ((unsigned int)(MsgCounter)));

    InitBacklog(pdx);
//    WRITE_IO_ULONG(CTRL_PORT, pdx->MaximumPacketSize == 64 ? FIFO_SET_FLU : FIFO_RES_FLU);

    if (LonStat == 0xFFFFFFFF || Stat == 0xFFFFFFFF || ByteCounter == 0xFFFFFFFF || MsgCounter == 0xFFFFFFFF)
        return FALSE;
    if (LonStat == Stat && Stat == ByteCounter && ByteCounter == MsgCounter)
        return FALSE;
#if 1 //defined DBG
    TimeStamp     = READ_IO_ULONG(CURR_RECV_TIME_PORT);
    Stat          = READ_IO_ULONG(STAT_PORT);
    LonStat       = READ_IO_ULONG(LON_STAT_PORT);
    ByteCounter   = READ_IO_ULONG(BYTE_COUNTER_PORT);
    MsgCounter    = READ_IO_ULONG(MSG_COUNTER_PORT);

    dbg("reset_neuron(2): TimeStamp=%08X LonStat=%08X Stat=%08X ByteCounter=%02X  MsgCounter=%02X",
        TimeStamp, LonStat, Stat, ByteCounter, MsgCounter);
#endif
    return TRUE;
    }


static int lvu_probe_create_usbdev(struct usb_interface *interface, const struct usb_device_id *id)
{
    struct usb_device *usbdev = interface_to_usbdev(interface);
    struct usb_host_interface *if_desc;
    PDEVICE_EXTENSION dev = NULL;
    struct usb_endpoint_descriptor *endpoint;
    int minor;
    int pipe;
    int i;
    int ret;
    char name[10];

    mutex_lock(&minor_table_mutex);
    for (minor = 0; minor < MAX_DEVICES; ++minor)
    {
        if (minor_table[minor] == NULL)
            break;
    }
    if (minor >= MAX_DEVICES)
    {
        err("Too many devices plugged in, can not handle this device.");
        ret = -ENODEV;
        goto exit;
    }

    dev = kzalloc(sizeof(DEVICE_EXTENSION), GFP_KERNEL);
    if (dev == NULL)
    {
        err("Out of memory");
        ret = -ENOMEM;
        goto exit;
    }
    minor_table[minor] = dev;

    dbg("got minor table entry [%i]", minor);
    mutex_init(&dev->mutex);

    dev->usbdev = usbdev;
    dev->interface = interface;
    dev->minor = minor;
    dev->remove_pending = FALSE;

    init_waitqueue_head(&dev->wr_wait);

    if (usbdev->descriptor.iProduct == 0)
    {
        dbg("Skipping rest of initialisation, waiting for disconnect");
        goto exitOK;
    }
#if 0
    InitNeuron(dev, FALSE);

    if (reset_neuron(dev, TRUE, TRUE, POWER_UP_RESET))
    {
        dev->HardwareWorking2 = TRUE;
    }

    dev->ServicePinTimeStamp = get_seconds();
#endif
    // set up the endpoint information
    // check out the endpoints
    if_desc = interface->cur_altsetting;  // 2.6.5
    for (i = 0; i < if_desc->desc.bNumEndpoints; ++i)
    {
        endpoint = &if_desc->endpoint[i].desc;
        dbg("bLength=%d "
            "bDescriptorType=%02X "
            "bEndpointAddress=%02X "
            "bmAttributes=%X "
            "wMaxPacketSize=%04X "
            "bInterval=%02X "
            "bRefresh=%02X "
            "bSynchAddress=%02X",
            endpoint->bLength,
            endpoint->bDescriptorType,
            endpoint->bEndpointAddress,
            endpoint->bmAttributes,
            endpoint->wMaxPacketSize,
            endpoint->bInterval,
            endpoint->bRefresh,
            endpoint->bSynchAddress);

        if (usb_endpoint_is_int_in(endpoint) && !dev->intr_in_buff)
        {
            /* we found a interrupt in endpoint */
            dbg("found interrupt endpoint(%d)", i);
            dev->intr_in_ep_addr = endpoint->bEndpointAddress;
            dev->bInterval = endpoint->bInterval;

            ret = CreateIntUrb(dev);
            if (ret)  goto error;
        }

        if (usb_endpoint_is_bulk_out(endpoint) && usb_endpoint_num(endpoint) == 2 && !dev->bulk_out_buffer)
            //            && !dev->write_urb)
        {
            /* we found a bulk out endpoint */
            dbg("found bulkout(%d)", i);
            dev->write_urb = usb_alloc_urb(0, GFP_KERNEL);
            if (!dev->write_urb)
            {
                err("No free (write) urbs available");
                ret = -ENOMEM;
                goto error;
            }
            pipe = usb_sndbulkpipe(usbdev, endpoint->bEndpointAddress);
            dev->bulk_out_size = usb_maxpacket(dev->usbdev, pipe, usb_pipeout(pipe));
            dbg("bulkpipe=%X bulk_out_size=%d", pipe, dev->bulk_out_size);
            dev->bulk_out_size = USB_WRITEBUFFER_SIZE;  // use maximum needed, because usb_submit_urb() does internal divide it

            dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;
            dev->bulk_out_buffer = kmalloc(dev->bulk_out_size, GFP_KERNEL);
            if (!dev->bulk_out_buffer)
            {
                err("Couldn't allocate bulk_out_buffer");
                ret = -ENOMEM;
                goto error;
            }
            usb_fill_bulk_urb(dev->write_urb, usbdev, pipe,
                dev->bulk_out_buffer, dev->bulk_out_size,
                lvu_write_bulk_callback, dev);
        }
    }

    dbg("lvu_probe_create_usbdev(): intr_in_buff=%p bulk_out_buffer=%p\n", dev->intr_in_buff, dev->bulk_out_buffer);
    if (!(dev->intr_in_buff && dev->bulk_out_buffer))
        goto error;

    sprintf(name, DRIVER_DEV_NAME "%d", dev->minor);

    /* let the user know what node this device is now attached to */
    pr_info(DRIVER_DESC " now attached to %s  major=%d, minor=%d\n", name, major, dev->minor);
exitOK:
    /* we can register the device now, as it is ready */
    usb_set_intfdata(interface, dev);

    /* we can register the device now, as it is ready */
    ret = usb_register_dev(interface, &lvu_class);
    if (ret)
    {
        /* something prevented us from registering this driver */
        err("Not able to get a minor for this device.");
        usb_set_intfdata(interface, NULL);
        goto error;
    }

    ret = 0;
    goto exit;

error:
    lvu_delete(dev);
    dev = NULL;

exit:

    mutex_unlock(&minor_table_mutex);
    dbg("%s() dev=%p minor=%d  return %d", __FUNCTION__, dev, minor, ret);
    return ret;
}


static int lvu_probe_create_device(PDEVICE_EXTENSION dev)
{
    int dev_offset;
    int ret;
    int SubdevNum;

    mutex_lock(&minor_table_mutex);

    dev_offset = dev->minor * MAX_DEVICENODES;

    cdev_init(&(dev->cdev), &lvu_fops);
    dev->cdev.owner = THIS_MODULE;
    dev->cdev.ops   = &lvu_fops;
    ret = cdev_add(&(dev->cdev), first_device + dev_offset, MAX_DEVICENODES);
/*
    dev->SubDevices = kzalloc(sizeof(struct device *) * MAX_DEVICENODES, GFP_KERNEL);
    if (dev->SubDevices == NULL)
    {
        err("Out of memory");
        cdev_del(&(dev->cdev));
        ret = -ENOMEM;
        goto error;
    }

    dev->pSubDev = kzalloc(sizeof(SUB_DEVICE_EXTENSION) * MAX_DEVICENODES, GFP_KERNEL);
    if (NULL == dev->pSubDev)
    {
        err("Out of memory");
        //kfree(dev->SubDevices);
        //dev->SubDevices = NULL;
        cdev_del(&(dev->cdev));
        ret = -ENOMEM;
        goto error;
    }
*/
    for (SubdevNum = 0; SubdevNum < MAX_DEVICENODES; SubdevNum++)
    {
        PSUB_DEVICE_EXTENSION subdev = &dev->pSubDev[SubdevNum];

        subdev->send_buffer_used = 0;

        qinit(&subdev->ReadBufferQueue, maxqueuecount);
        init_waitqueue_head(&subdev->rd_wait);

        dev->SubDevices[SubdevNum] = DEVICE_CREATE(lvu_dev_class, NULL, first_device + dev_offset + SubdevNum, NULL, "lvu_%i", SubdevNum + dev_offset);

        subdev->udev   = dev;
        subdev->SubdevNum = SubdevNum;
#if USE_PROC
        //sprintf(subdev->DevStats.ProcDevName, "lvu_%i", SubdevNum + dev_offset);
		sprintf(subdev->DevStats.ProcDevName, PROC_DEV_NAME "_%d", SubdevNum + dev_offset);
        if (HasEasylonProcDir)
        {
            dbg("proc_create_data(%s, %p):\n", subdev->DevStats.ProcDevName, subdev);
            proc_create_data(subdev->DevStats.ProcDevName, 0, NULL, &lonusb_proc_fops_subdev, subdev);
        }
#endif
    }

//error:
    mutex_unlock(&minor_table_mutex);
    return ret;
}


/**
 *      lvu_probe
 *
 *      Called by the usb core when a new device is connected that it thinks
 *      this driver might be interested in.
 */
static int lvu_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
    struct usb_device *usbdev = interface_to_usbdev(interface);
    PDEVICE_EXTENSION pdx = NULL;
    int ret = -ENODEV;
    LPX_CARD LpxCard = LPX_INVALID;
    UINT RamSize = RAMSIZE;

    dbg("%s(): found usb idVendor=%#x, idProduct=%#x iProduct=%#x interface_num:%d",
        __FUNCTION__,
        usbdev->descriptor.idVendor,
        usbdev->descriptor.idProduct,
        usbdev->descriptor.iProduct,
        interface->cur_altsetting->desc.bInterfaceNumber);  // 2.6.5

    // See if the device offered us matches what we can accept
    if (cpu_to_le16(usbdev->descriptor.idVendor) != GESYTEC_USB_VID)
    {
        dbg("%s(): no match", __FUNCTION__);
        return -ENODEV;
    }

    switch (cpu_to_le16(usbdev->descriptor.idProduct))
    {
    case GESYTEC_LVU_PID:
    case GESYTEC_LVX2_PID:
    case GESYTEC_LVX3_PID:
        LpxCard = LVU001;
        break;
    case GESYTEC_GDS_PID:
    case GESYTEC_GDS_LATTICE_PID:
    	if (cpu_to_le16(usbdev->descriptor.bcdDevice) != GESYTEC_GDS203_REV)
		{
			LpxCard = GDS201;
			RamSize = RAMSIZE_GDS;
			break;
		}
    	// for GESYTEC_GDS203_REV fall through
    case GESYTEC_LVU2_PID:
    case GESYTEC_LVU2_LATTICE_PID:
        LpxCard = LVU2;
        RamSize = RAMSIZE_GDS;
        break;
    default:
        dbg("%s(): no match", __FUNCTION__);
        return -ENODEV;
    }

    dbg("%s(): Test2 LpxCard=%d ID=%i", __FUNCTION__, LpxCard, usbdev->descriptor.iProduct);
    if (usbdev->descriptor.iProduct == 0)
    {
DownloadFirmware:
        dbg("ezusb_load_firmware(%p):", usbdev);
        dbg("ezusb_load_firmware1: START");
        ret = lonusb_dl_formware(usbdev);
        dbg("ezusb_load_firmware1: ret=%d", ret);
        dbg("ezusb_load_firmware(%p)=%d return", usbdev, ret);

        //ret = lvu_probe_create_usbdev(interface, id);
        ret = -ENODEV;
    }
    else
    {
        ret = lvu_probe_create_usbdev(interface, id);

        if (0 == ret)
        {
            pdx = usb_get_intfdata(interface);

            if (pdx)
            {
                NTSTATUS  status = 0;
                BYTE buffer[8];
                ULONG ret = Ezusb_VendorRequest(pdx->usbdev,
                    //Req          Value   Index   Len             1=in
                    VR_FW_VERSION, 0x0000, 0x0000, sizeof(buffer), 1, buffer);
                DrvPrint((_t("Ezusb_VendorRequest(VR_FW_VERSION)=%X buffer=%02X %02X %02X %02X %02X %02X %02X %02X  Done=%02X\n"), ret,
                    buffer[0], buffer[1], buffer[2], buffer[3],
                    buffer[4], buffer[5], buffer[6], buffer[7],
                    buffer[3]));
                if (ret < 4 )
                    // || VendorRequest.buffer[0] != VersionInfo[8]  // VER_FILEVERSION_MAJOR
                    // || VendorRequest.buffer[1] != VersionInfo[9]) // VER_FILEVERSION_MINOR)
                    goto DownloadFirmware;
                pdx->RevisionID = buffer[2];

                pdx->LpxCard = LpxCard;
                if (LpxCard == LVU001)
                    status = XilinxFpgaCheckAndLoad(pdx);  // LpxCard, usbdev); //, FpgaFile1, sizeof (FpgaFile1));
                else if (LpxCard == LVU2)
                    status = LatticeFpgaCheckAndLoad(pdx);

                InitNeuron(pdx, FALSE);

                if (status == 0)
                {
                    int MaxLoop = 2;  //1000;
                    while (!reset_neuron(pdx, FALSE, TRUE, POWER_UP_RESET))
                    {
                        udelay(50);
                        if (--MaxLoop <= 0)  break;
                    }

                    if (/*doreset &&*/ !RamTest(pdx, 0, RamSize -1))
                    {
                        pdx->error_log   = MEMORY_ALLOC_FAILURE;    // RAM memory allocation failure
                        status = -1; // STATUS_PARITY_ERROR;
                    }
                    else
                    {
                        if (reset_neuron(pdx, TRUE, TRUE, POWER_UP_RESET))
                        {
                            pdx->HardwareWorking2 = TRUE;
                            //EnableIrq(pdx, TRUE);
                        }
                        else
                        {
                            DrvPrint(("reset_neuron()=???\n"));
                            pdx->error_log   = APPL_SIGNATURE_BAD;  // Application image inconsistency
                            status = -1;  // STATUS_DRIVER_INTERNAL_ERROR;
                        }
                    }

                }
                else
                    pdx->error_log   = SELF_TEST_FAILED;    // Self test failed
/*
                if (reset_neuron(pdx, TRUE, TRUE, POWER_UP_RESET))
                {
                pdx->HardwareWorking2 = TRUE;
                }
*/
                pdx->ServicePinTimeStamp = get_seconds();

                ret = lvu_probe_create_device(pdx);
#if USE_PROC
                {
#ifdef CONFIG_USB_DYNAMIC_MINORS
                    // see minor_base at http://www.makelinux.net/ldd3/chp-13-sect-4
                    int minor = interface->minor;
#else
                    int minor = interface->minor - lvu_class.minor_base;
#endif
                    sprintf(pdx->DevStats.ProcDevName, PROC_DEV_NAME "%d" /*lvu_class.name*/, minor);
                    if (HasEasylonProcDir)
                    {
                        dbg("proc_create_data(%s, %p):\n", pdx->DevStats.ProcDevName, pdx);
                        proc_create_data(pdx->DevStats.ProcDevName, 0, NULL, &lonusb_proc_fops_dev, pdx);
                    }

                    ProcStats.probe_cnt++ ;
                }
#endif
            }
        }
    }

    return ret;
}


static void lvu_disconnect_usbdev(struct usb_interface *interface, PDEVICE_EXTENSION dev)
{
    dbg("lvu_disconnect_usbdev: entered");

    mutex_lock(&dev->mutex);

    dev->remove_pending = TRUE; // Now all data will be ignored

    wake_up_interruptible(&dev->wr_wait);

    dbg("lvu_disconnect_usbdev: enter loop over char SubDevices");

    // if the device is not opened, then we clean up right now
    if (0 == dev->sub_open_count)
    {
        mutex_unlock(&dev->mutex);
        lvu_delete(dev);
    }
    else
    {
        // shutdown any bulk writes that might be going on
        dump_urb(dev->read_urb);
        usb_kill_urb(dev->read_urb);
        dump_urb(dev->read_urb);
        usb_kill_urb(dev->write_urb);

        dev->usbdev = NULL;
        mutex_unlock(&dev->mutex);
    }

    usb_set_intfdata(interface, NULL);
    // give back our minor
    usb_deregister_dev(interface, &lvu_class);
}


static void lvu_disconnect_device(struct usb_interface *interface, PDEVICE_EXTENSION dev)
{
    int SubdevNum;
    int dev_offset = dev->minor * MAX_DEVICENODES;

    dbg("lvu_disconnect_device: entered");

    mutex_lock(&dev->mutex);

    dev->remove_pending = TRUE; /* Now all data will be ignored */

    dbg("lvu_disconnect_device: vor wake_up()");
    wake_up_interruptible(&dev->wr_wait);

    dbg("lvu_disconnect_device: enter loop over char SubDevices");

    //if (NULL != dev->pSubDev)
    {
        for (SubdevNum = 0; SubdevNum < MAX_DEVICENODES; SubdevNum++)
        {
            PSUB_DEVICE_EXTENSION subdev = &dev->pSubDev[SubdevNum];

            dbg("lvu_disconnect_device: kill device %d", SubdevNum);
            wake_up_interruptible(&subdev->rd_wait);

            qexit(&subdev->ReadBufferQueue);

            device_destroy(lvu_dev_class, first_device + dev_offset + SubdevNum);
#if USE_PROC
            if (HasEasylonProcDir)
                remove_proc_entry(subdev->DevStats.ProcDevName, NULL);
#endif
        }

        //printk(KERN_DEBUG "lvu_disconnect_device(dev=%p): pSubDev=%p\n", dev, dev->pSubDev);
        //kfree(dev->pSubDev);
        //dev->pSubDev = NULL;
    }

    dbg("kill kernel device nodes");
/*
    if (NULL != dev->SubDevices)
    {
        kfree(dev->SubDevices);
        dev->SubDevices = NULL;
    }
*/
    cdev_del(&(dev->cdev));

    mutex_unlock(&dev->mutex);
}


/**
 *      lvu_disconnect
 *
 *      Called by the usb core when the device is removed from the system.
 */
static void lvu_disconnect(struct usb_interface *interface)
{
    PDEVICE_EXTENSION dev = usb_get_intfdata(interface);
    dbg("%s(): dev=%p", __FUNCTION__, dev);

    if (dev)
    {
        mutex_lock(&minor_table_mutex);

        //if (NULL != dev->pSubDev)
        {
            lvu_disconnect_device(interface, dev);
        }

        lvu_disconnect_usbdev(interface, dev);

        mutex_unlock(&minor_table_mutex);

#if USE_PROC
        if (HasEasylonProcDir)
            remove_proc_entry(dev->DevStats.ProcDevName, NULL);

        ProcStats.disconnect_cnt++;
#endif
    }

    pr_info(DRIVER_DESC " " DRIVER_DEV_NAME "%d now disconnected\n", dev->minor);
}


/**
 *      lvu_init
 */
static int __init lvu_init(void)
    {
    int result;
    int ret;
    dev_t from;

#if USE_PROC
    if (CreateEasylonProcDir())
        proc_create_data(PROC_DEV_NAME, 0, NULL, &lonusb_proc_fops_all, NULL);
#endif

    /* register this driver with the USB subsystem */

    lvu_dev_class = class_create(THIS_MODULE, "lvu");

    Major = register_chrdev(0, DRIVER_DEV_NAME, &lvu_fops);
    if (Major < 0)
    {
        err(" Major number allocation is failed \n");
        return Major;
    }
    unregister_chrdev(Major, DRIVER_DEV_NAME);

    from = MKDEV(Major, 0);

    ret = register_chrdev_region(from, MAX_DEVICETABLE, DRIVER_DEV_NAME);

    if (0 == ret)
      {
        dbg("registered for character devices for major %d", Major);

        first_device = from;
      }
    else
      {
        err("register_chrdev_region(%08X, 4, %s) => %i", from, DRIVER_DEV_NAME, ret);
      }

    dbg(" The Major number of the device is %d \n", major);

    result = usb_register(&lvu_driver);
    if (result < 0) {
        err("usb_register failed for the "__FILE__" driver. Error number %d",
            result);
        return -1;
        }

    pr_info(DRIVER_DESC " " DRIVER_VERSION " result=%d\n", result);
    return 0;
    }


/**
 *      lvu_exit
 */
static void __exit lvu_cleanup(void)
{
#if USE_PROC
	if (HasEasylonProcDir)
		remove_proc_entry(PROC_DEV_NAME, NULL);
    RemoveEasylonProcDir();
#endif
    /* deregister this driver with the USB subsystem */
    usb_deregister(&lvu_driver);

    if (0 != first_device)
    {
        unregister_chrdev_region(first_device, MAX_DEVICETABLE);
    }

    class_destroy(lvu_dev_class);

    dbg("The Major number is released successfully \n");
}


module_exit(lvu_cleanup);
module_init(lvu_init);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

MODULE_INFO(supported, "external");
