/*++

Module Name:

    fpga.cpp

Abstract:

    Spartan IIe fpga loader for Gesytec LVP

Environment:

    kernel mode only

Revision History:

    19.04.04 VS: Created

--*/



///////////////////////////////////////////////////////////////////////////////

//#pragma PAGEDCODE

#define FIRMWARE_XILINX	    "gesytec/lvu_1.bit"
#define FIRMWARE_LATTICE	"gesytec/gds_2.fw"

MODULE_FIRMWARE(FIRMWARE_XILINX);
MODULE_FIRMWARE(FIRMWARE_LATTICE);


//#include "spartan2e.h"
//UCHAR FpgaDummySpace[128] = {1,2,3,4,5,6,7,8};

#define CHUNK_SIZE 64

// .bit-file header format
struct xbitfile {
	USHORT	hdr_size;		// +0x00
	UCHAR	hdr_chars[9];	// +0x02
	USHORT	version;		// +0x0B

	UCHAR	*sourcefile;	// 'a' tag
	UCHAR	*devicetype;	// 'b' tag
	UCHAR	*date;			// 'c' tag
	UCHAR	*time;			// 'd' tag
	int32_t  size;			// 'e' tag
	UCHAR	*data;			//
    };

// struct containing all information from .bit-file
static struct xbitfile bitfile;

/*
VOID _inline udelay(ULONG MicroSeconds)
    {
#if 0
    if (KeGetCurrentIrql() == PASSIVE_LEVEL)
        {
        LARGE_INTEGER  Interval;
        Interval.QuadPart = -((LONGLONG)MicroSeconds *10);
        KeDelayExecutionThread(KernelMode, FALSE, &Interval);
        }
    else
#endif
        KeStallExecutionProcessor(MicroSeconds);
    }
*/

#if 0
// swap 16-Bit Word from LE to BE and vice versa
static inline USHORT _swapw(volatile USHORT v)
    {
    return ((v << 8) | (v >> 8));
    }


// swap 32-Bit Word from LE to BE and vice versa
static inline ULONG _swapl(volatile ULONG v)
    {
    return ((v << 24) | ((v & 0xff00) << 8) | ((v & 0xff0000) >> 8) | (v >> 24));
    }
#else
	#define _swapw(v) be16_to_cpu(v)
	#define _swapl(v) be32_to_cpu(v)
#endif

// Parse header of bit-file and fill struct.

static int parse_header(UCHAR *image, int image_size)
     {
     USHORT *shortptr;
     ULONG  *longptr;
     UCHAR  *charptr;

     shortptr = (PUSHORT)&image[0];
     bitfile.hdr_size = _swapw(*shortptr);
     memcpy(bitfile.hdr_chars, &image[2], 9);
     shortptr = (PUSHORT)&image[11];
     bitfile.version = _swapw(*shortptr);

     charptr = image + 13;
     while (charptr < (image + image_size))
         {
         longptr = (PULONG)(charptr+1);
         shortptr = (PUSHORT)longptr;
         switch (*charptr++) {
             case 'a':
                 bitfile.sourcefile = charptr + 2;
                 charptr += _swapw(*shortptr) + 2;
                 break;
             case 'b':
                 bitfile.devicetype = charptr + 2;
                 charptr += _swapw(*shortptr) + 2;
                 break;
             case 'c':
                 bitfile.date = charptr + 2;
                 charptr += _swapw(*shortptr) + 2;
                 break;
             case 'd':
                 bitfile.time = charptr + 2;
                 charptr += _swapw(*shortptr) + 2;
                 break;
             case 'e':
                 bitfile.size = _swapl(*longptr);
                 bitfile.data = charptr + 4;
                 charptr += 4 + bitfile.size;
                 break;
             default:
                 dbg("Unknown token: %c", *(charptr-1));
                 break;
             }
         }
     return 0;
     }


// Plot full information about the programmed bitfile into the specified string.

static BOOLEAN plot_header(void)
     {
     BOOLEAN ret = FALSE;
     static UCHAR expected_hdr_chars[] = { 0x0F, 0xF0, 0x0F, 0xF0, 0x0F, 0xF0, 0x0F, 0xF0, 0x00 };
     int i;

     dbg("  Header:\t");
     if (bitfile.hdr_size != 9)
         dbg("invalid (size is %d)", bitfile.hdr_size);
     else {
         for (i=0; i<9; i++)
             {
             if (bitfile.hdr_chars[i] != expected_hdr_chars[i])
                 break;
             }
         if (i == 9)
             {
             dbg("ok");
             ret = TRUE;
             }
         else
             dbg("invalid");
         }
     dbg("  Version:\t%d", bitfile.version);
     dbg("  Source file:\t%s", bitfile.sourcefile);
     dbg("  Device type:\t%s", bitfile.devicetype);
     dbg("  Date:\t\t%s", bitfile.date);
     dbg("  Time:\t\t%s", bitfile.time);
     dbg("  Size:\t\t%d bytes", bitfile.size);
     dbg("  LoadType:\t\t%02X=%s", bitfile.data[0x1D],
         bitfile.data[0x1D] == 0x80 ? "Parallel" : bitfile.data[0x1D] == 0xA0 ? "JTAG!!!" : "???");
     if (bitfile.data[0x1D] != 0x80)  ret = FALSE;
     return ret;
     }




NTSTATUS LatticeFpgaLoad(PDEVICE_EXTENSION pdx, XO2_JEDEC_t *pProgJED)  //, UCHAR *image, int image_size)
    {
    NTSTATUS ntStatus = -1;  // STATUS_ADAPTER_HARDWARE_ERROR;
    if (!(/*pdx->LpxCard == GDS201 ||*/ /*pdx->LpxCard == GDS203 ||*/ pdx->LpxCard == LVU2 /*|| pdx->LpxCard == LVX4*/))  return 0;  // STATUS_SUCCESS;
    // Download for GDS201 no longer supported

    DrvPrint((_t("LatticeFpgaLoad() LpxCard=%d:\n"), pdx->LpxCard));
#if 01
        {
        ULONG ret = XO2ECAcmd_openCfgIF(pdx, TRANSPARENT_MODE);
        if (!ret)  return -1;  // STATUS_ADAPTER_HARDWARE_ERROR;

        ret = XO2ECA_apiGetDevType(pdx);
        DrvPrint((_t("XO2ECA_apiGetDevType()=%d devType=%08X\n"), ret, pdx->devType));
        if (!ret)  return -1;  // STATUS_ADAPTER_HARDWARE_ERROR;

/*
        unsigned int  Stat = 0;
        ret = XO2ECAcmd_readStatusReg(pdx, &Stat);
        ___DrvPrint((_t("XO2ECAcmd_readStatusReg()=%d Stat=%08X\n"), ret, Stat));
        if (!ret)  return STATUS_ADAPTER_HARDWARE_ERROR;
*/
        XO2ECAcmd_closeCfgIF(pdx);
        XO2ECAcmd_Bypass(pdx);


        ret = XO2ECA_apiProgram(pdx, pProgJED, 
            //XO2ECA_PROGRAM_OFFLINE | 
            XO2ECA_PROGRAM_TRANSPARENT |
            XO2ECA_PROGRAM | XO2ECA_VERIFY | 
            XO2ECA_ERASE_PROG_CFG | XO2ECA_ERASE_PROG_UFM);  // | XO2ECA_ERASE_PROG_FEATROW);
        DrvPrint((_t("XO2ECA_apiProgram()=%d\n"), ret));
        if (ret == OK)  ntStatus = 0;  // STATUS_SUCCESS;

        ezusb_set_reset(pdx, 1);  // Reset firmware because of some problems
        ezusb_set_reset(pdx, 0);  // after starting with blank or corrupt FPGA
        mSecDelay(50);  // wait 50 ms, because of some startup delays in the firmware
        }
    return ntStatus;
#else
    return STATUS_SUCCESS;
#endif
    }

NTSTATUS LatticeFpgaCheckAndLoad(PDEVICE_EXTENSION pdx)
    {
    DrvPrint((_t("LatticeFpgaCheckAndLoad() LpxCard=%d:\n"), pdx->LpxCard));
        {
        XO2_JEDEC_t ProgJED;
        PUCHAR image = NULL;
        int image_size = 0, used_size = 0;
        const struct firmware *lattice_fw;
        ULONG ret;
        NTSTATUS ntStatus = 0;  // STATUS_SUCCESS;
        BYTE buffer[8];
        dbg("LatticeFpgaCheckAndLoad() RevisionID=%d", pdx->RevisionID);
        switch (pdx->RevisionID)
            {
            case 1:
                {
                int err = request_firmware(&lattice_fw, FIRMWARE_LATTICE, &pdx->usbdev->dev);
                if (err < 0) 
                {
                    dev_err(&pdx->usbdev->dev, "Failed to load %s: %d\n", FIRMWARE_LATTICE, err);
                    return err;
                }
                image = (PUCHAR)lattice_fw->data;
                image_size = lattice_fw->size;
                if (image_size < sizeof(ProgJED))
                    {
                    ntStatus = -1;
                    goto err_fw;
                    }
                memcpy(&ProgJED, image, sizeof(ProgJED));
                used_size = FIELD_OFFSET(XO2_JEDEC_t, pCfgData) + ProgJED.CfgDataSize + ProgJED.UFMDataSize + sizeof(XO2FeatureRow_t);
                dbg("LatticeFpgaCheckAndLoad(): image_size=%d used_size=%d", image_size, used_size);
                if (image_size != used_size)
                    {
                    ntStatus = -1;
                    goto err_fw;
                    }
                ProgJED.pCfgData = image + FIELD_OFFSET(XO2_JEDEC_t, pCfgData);
                ProgJED.pUFMData = ProgJED.pCfgData + ProgJED.CfgDataSize;
                ProgJED.pFeatureRow = (XO2FeatureRow_t *)(ProgJED.pUFMData + ProgJED.UFMDataSize);
                XO2ECA_apiJEDECinfo(pdx, &ProgJED);

                //image = FpgaFile1;
                //image_size = sizeof(FpgaFile1);
                //dev_err(&pdx->usbdev->dev, "request_firmware(%s): size=%d image_size=%d\n", FIRMWARE_LATTICE, lattice_fw->size, image_size);
                break;
                }
            case 0:
            default:
                return -1; //STATUS_RESOURCE_TYPE_NOT_FOUND;  // STATUS_INVALID_IMAGE_FORMAT
            }
            
        ret = Ezusb_VendorRequest(pdx->usbdev, 
            //Req          Value   Index   Len             1=in
            VR_FW_VERSION, 0x0000, 0x0000, sizeof(buffer), 1, buffer);
        DrvPrint ((_t("Ezusb_VendorRequest(VR_FW_VERSION)=%X buffer=%02X %02X %02X %02X %02X %02X %02X %02X  Done=%02X\n"), ret,
            buffer[0], buffer[1], buffer[2], buffer[3],
        buffer[4], buffer[5], buffer[6], buffer[7],
        buffer[3]));
        if (ret < 4 || !buffer[3])  // FPGA_DONE
            {
            pdx->RevisionID = buffer[2];
            ntStatus = LatticeFpgaLoad(pdx, &ProgJED);
            }

    ret = XO2ECAcmd_openCfgIF(pdx, TRANSPARENT_MODE);
            {
            unsigned int  Stat = 0;
            ret = XO2ECAcmd_readStatusReg(pdx, &Stat);
            DrvPrint((_t("XO2ECAcmd_readStatusReg()=%d Stat=%08X\n"), ret, Stat));
            if (!ret)  return -1;  // STATUS_ADAPTER_HARDWARE_ERROR;
	// Verify that DONE bit is definitely set and not FAIL or BUSY
    if ((Stat & 0x3100) != 0x100)
                goto DownloadLattice;
            }

        ret = XO2ECAcmd_readUserCode(pdx, &pdx->UserCode);
        DrvPrint ((_t("XO2ECAcmd_readUserCode()=%d UserCode=%08X ProgJED.UserCode=%08X\n"), ret, pdx->UserCode, ProgJED.UserCode));
    XO2ECAcmd_closeCfgIF(pdx);
    XO2ECAcmd_Bypass(pdx);

        if (!ret)  return -1;  // STATUS_ADAPTER_HARDWARE_ERROR;
        if (pdx->UserCode != ProgJED.UserCode)
            {
DownloadLattice:
            ntStatus = LatticeFpgaLoad(pdx, &ProgJED);
            }
err_fw:
        release_firmware(lattice_fw);
        return ntStatus;
        }
    }							// FpgaCheckAndLoad


static int XilinxFpgaLoad(PDEVICE_EXTENSION pdx);

static int XilinxFpgaCheckAndLoad(PDEVICE_EXTENSION pdx)
    {
    int status = 0;
    BYTE buffer[8];
    ULONG ret;
    memset (buffer, 0x00, sizeof(buffer));
    ret = Ezusb_VendorRequest(pdx->usbdev,
        //Req          Value   Index   Len             1=in
        VR_FW_VERSION, 0x0000, 0x0000, sizeof(buffer), 1, buffer);
    dbg("XilinxFpgaCheckAndLoad(1): Ezusb_VendorRequest(VR_FW_VERSION)=%X buffer=%02X %02X %02X %02X %02X %02X %02X %02X", (unsigned int) (ret),
        buffer[0], buffer[1], buffer[2], buffer[3],
        buffer[4], buffer[5], buffer[6], buffer[7]);
    if (ret < 4 || !buffer[3])  // FPGA_DONE
        {
           pdx->RevisionID = buffer[2];
           dbg("XilinxFpgaCheckAndLoad(2): RevisionID=%02X", buffer[2]);
           status = XilinxFpgaLoad(pdx);  // pdx->LpxCard, pdx->usbdev, buffer[2]);
        }

    ret = Ezusb_VendorRequest(pdx->usbdev,
        //Req          Value   Index   Len             1=in
        VR_FW_VERSION, 0x0000, 0x0000, sizeof(buffer), 1, buffer);
            dbg("XilinxFpgaCheckAndLoad(3): Ezusb_VendorRequest(VR_FW_VERSION)=%X buffer=%02X %02X %02X %02X %02X %02X %02X %02X", (unsigned int) (ret),
        buffer[0], buffer[1], buffer[2], buffer[3],
        buffer[4], buffer[5], buffer[6], buffer[7]);

    return status;

    }							// XilinxFpgaCheckAndLoad


static int XilinxFpgaLoad(PDEVICE_EXTENSION pdx)
    {
    //int status = -1; //STATUS_ADAPTER_HARDWARE_ERROR;
    PUCHAR image = NULL;
    int image_size = 0;
    PUCHAR ptr;
    int chunkCount;
    int i;
    ULONG ret = -1;
	const struct firmware *xilinx_fw;

    //if (image == NULL)
        {
        dbg("XilinxFpgaLoad() RevisionID=%d", pdx->RevisionID);
        switch (pdx->RevisionID)
            {
            case 1:
                {
                int err = request_firmware(&xilinx_fw, FIRMWARE_XILINX, &pdx->usbdev->dev);
                if (err < 0) 
                {
                    dev_err(&pdx->usbdev->dev, "Failed to load %s: %d\n", FIRMWARE_XILINX, err);
                    return err;
                }
                image = (PUCHAR)xilinx_fw->data;
                image_size = xilinx_fw->size;

                //image = FpgaFile1;
                //image_size = sizeof(FpgaFile1);
                //dev_err(&pdx->usbdev->dev, "request_firmware(%s): size=%d image_size=%d\n", FIRMWARE_XILINX, xilinx_fw->size, image_size);
                break;
                }
            case 0:
            default:
                return -1; //STATUS_RESOURCE_TYPE_NOT_FOUND;  // STATUS_INVALID_IMAGE_FORMAT
            }
        }
    parse_header(image, image_size);
    if (!plot_header())
        goto err_fw; //STATUS_RESOURCE_TYPE_NOT_FOUND;  // STATUS_INVALID_IMAGE_FORMAT

    ptr = bitfile.data;
/*
    PURB urb = (PURB)ExAllocatePool(NonPagedPool, sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST));
    if (!urb)  return STATUS_NO_MEMORY;
*/

    chunkCount = ((bitfile.size + CHUNK_SIZE - 1) / CHUNK_SIZE);
    //
    // The download will be split into CHUNK_SIZE pieces and
    // downloaded with multiple setup transfers.  For the Rev B parts
    // CHUNK_SIZE should not exceed 64 bytes, as larger transfers can
    // result in data corruption when other USB devices are present.
    //
    for (i = 0; i < chunkCount; i++)
        {
        /*RtlZeroMemory(urb,sizeof(struct  _URB_CONTROL_VENDOR_OR_CLASS_REQUEST));

        urb->UrbHeader.Length = sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST);
        urb->UrbHeader.Function = URB_FUNCTION_VENDOR_DEVICE;

        if ((i == (chunkCount - 1)) &&  (bitfile.size % CHUNK_SIZE))
            {
            urb->UrbControlVendorClassRequest.TransferBufferLength =
                (bitfile.size % CHUNK_SIZE);
            urb->UrbControlVendorClassRequest.Value = 2;  // last
            }
        else
            {
            urb->UrbControlVendorClassRequest.TransferBufferLength = CHUNK_SIZE;
            urb->UrbControlVendorClassRequest.Value = (i==0) ? 0 : 1;  // 0 = first
            }

        urb->UrbControlVendorClassRequest.TransferBuffer = ptr;
        urb->UrbControlVendorClassRequest.TransferBufferMDL = NULL;
        urb->UrbControlVendorClassRequest.Request = VR_FPGA;
        urb->UrbControlVendorClassRequest.Index = 0;

        DrvPrint (("Function=%02X TransferFlags=%02X TransferBufferLength=%02X  Request=%02X Value=%02X Index=%02X\n",
            urb->UrbHeader.Function, urb->UrbControlVendorClassRequest.TransferFlags,
            urb->UrbControlVendorClassRequest.TransferBufferLength,
            urb->UrbControlVendorClassRequest.Request,
            urb->UrbControlVendorClassRequest.Value,
            urb->UrbControlVendorClassRequest.Index
            ));
        ntStatus = Ezusb_CallUSBD(pdx, urb);*/

        if ((i == (chunkCount - 1)) &&  (bitfile.size % CHUNK_SIZE))
        {
          ret = Ezusb_VendorRequest(pdx->usbdev,
                //Req          Value   Index   Len             1=in
                VR_FPGA, 2, 0x0000, (bitfile.size % CHUNK_SIZE), 0, ptr);
        }
        else
        {
          ret = Ezusb_VendorRequest(pdx->usbdev,
                //Req          Value   Index   Len             1=in
                VR_FPGA, (i==0) ? 0 : 1, 0x0000, CHUNK_SIZE, 0, ptr);
        }

        if (ret < 0)
            break;

        ptr += CHUNK_SIZE;
        }

err_fw:
	release_firmware(xilinx_fw);
    /*ExFreePool(urb);*/
    return ret < 0;
    }							// XilinxFpgaLoad


///////////////////////////////////////////////////////////////////////////////

UCHAR Peek(PDEVICE_EXTENSION pdx, int Addr)
    {
    WRITE_IO_ULONG(SET_RAM_ADDR_PORT, Addr);
    return (UCHAR) READ_IO_ULONG(RAM_TEST_DATA_PORT);
    }

VOID Poke(PDEVICE_EXTENSION pdx, int Addr, UCHAR Data)
    {
    WRITE_IO_ULONG(SET_RAM_ADDR_PORT, Addr);
    WRITE_IO_ULONG(RAM_TEST_DATA_PORT, Data);
    }

BOOLEAN RamTest(PDEVICE_EXTENSION pdx, UINT first, UINT last)
    {
//    KIRQL               oldIrql;
//    KeAcquireSpinLock(&pdx->NeuronSpinLock, &oldIrql);  // -> DISPATCH_LEVEL

//    __try
        {
        UCHAR   i, rval, wval;
        UINT aline;

        WRITE_IO_ULONG(CTRL_PORT, NLON_RESET | HOST_RESET);
        udelay(10);
        Printf(("RamTest():\n"));
        WRITE_IO_ULONG(CTRL_PORT, RCV_DIS_STB);
        udelay(10);

        Printf(("Datenleitungen :"));
        for(i=0, wval=1; i<8; i++, wval<<=1)
            {
            Poke(pdx, first, wval);
            Poke(pdx, first+1, ~wval);
            if( ( rval = Peek(pdx, first) ) != wval )
                {
                Printf(( "\nData Line Error: D %X\n", i));
                Printf(( "Fault Address: %04X  ", first));
                Printf(( "Write: %02X  Read: %02X\n", wval, rval));
                return FALSE;
                }
            if( ( rval = Peek(pdx, first+1) ) != (UCHAR)(~wval) )
                {
                Printf(( "\nData Line Error: D %X\n", i));
                Printf(( "Fault Address: %04X  ", first));
                Printf(( "Write: %02X  Read: %02X\n", (UCHAR)(~wval), rval));
                return FALSE;
                }
            }
        for(i=0, wval=~1; i<8; i++, wval<<=1)
            {
            Poke(pdx, first, wval);
            Poke(pdx, first + 1, ~wval);
            if( ( rval = Peek(pdx, first) ) != wval )
                {
                Printf(( "\nData Line Error: D %X\n", i));
                Printf(( "Fault Address: %04X  ", first));
                Printf(( "Write: %02X  Read: %02X\n", wval, rval));
                return FALSE;
                }
            if( ( rval = Peek(pdx, first+1) ) != (UCHAR)(~wval) )
                {
                Printf(( "\nData Line Error: D %X\n", i));
                Printf(( "Fault Address: %04X  ", first));
                Printf(( "Write: %02X  Read: %02X\n", (UCHAR)(~wval), rval));
                return FALSE;
                }
            }
        Printf(("  OK\n"));

        Printf(("Adressleitungen:"));

        wval = 0xA5;
        for(i=0, aline=1; aline<last; i++, aline<<=1)
            {
            Poke(pdx, aline, wval);
            Poke(pdx, first, ~wval);
            if( ( rval = Peek(pdx, aline) ) != wval )
                {
                Printf(( "\nAddress Line Error: A %X\n", i));
                Printf(( "Fault Address: %04X  ", aline));
                Printf(( "Write: %02X  Read: %02X\n", wval, rval));
                return FALSE;
                }
            if( ( rval = Peek(pdx, first) ) != (UCHAR)(~wval) )
                {
                Printf(( "\nAddress Line Error: A %X\n", i));
                Printf(( "Fault Address: %04X  ", first));
                Printf(( "Write: %02X  Read: %02X\n", (UCHAR)(~wval), rval));
                return FALSE;
                }
            }

        wval = 0x5A;
        for(i=0, aline=(~1) & last; aline<last; i++, aline = ((aline <<1) | 1) & last)
            {
            Poke(pdx, aline, wval);
            Poke(pdx, last, ~wval);
            if( ( rval = Peek(pdx, aline) ) != wval )
                {
                Printf(( "\nAddress Line Error: A %X\n", i));
                Printf(( "Fault Address: %04X  ", aline));
                Printf(( "Write: %02X  Read: %02X\n", wval, rval));
                return FALSE;
                }
            if( ( rval = Peek(pdx, last) ) != (UCHAR)(~wval) )
                {
                Printf(( "\nAddress Line Error: A %X\n", i));
                Printf(( "Fault Address: %04X  ", last));
                Printf(( "Write: %02X  Read: %02X\n", (UCHAR)(~wval), rval));
                return FALSE;
                }
            }
        Printf(("  OK\n"));

#if 0
        Printf(("Pattern Test   :"));
        UINT Addr;
        for (Addr=0; Addr <= last; Addr++)
            Poke(pdx, Addr, Addr ^ (Addr >> 8) ^ 0xA5);
        for (Addr=0; Addr <= last; Addr++)
            {
            rval = Peek(pdx, Addr);
            wval = Addr ^ (Addr >> 8) ^ 0xA5;
            if (rval != wval)
                {
                Printf(("\n[%04X] Write: %02X  Read: %02X\n", Addr, wval, rval));
                return FALSE;
                }
            }
        Printf(("  OK\n"));
#endif
//        } __finally
//        {
//        KeReleaseSpinLock(&pdx->NeuronSpinLock, oldIrql);
        }
    return TRUE;
    }
