/*****************************************************************************/
/*
 * Linux driver for Easylon USB VNI Interfaces
 *
 * Copyright (c)2015 Gesytec GmbH, Volker Schober (info@gesytec.de)
 * Licensend for use with Easylon USB Interfaces by Gesytec
 *
 * Please compile lvu.c
 *
 */
 /*****************************************************************************/

#ifndef __lonusb_h__
#define __lonusb_h__

//#define DEBUG  // for dev_dbg()
#define DEBUG_ECA 0


#include <linux/firmware.h>
#include <linux/kernel.h>

#include <linux/errno.h>
#include <linux/poll.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/completion.h>

#include <linux/usb.h>
#include <linux/version.h>
#include <linux/cdev.h>

#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#include "sysdep.h"
#include "GesytecIDs.h"
#include "dynbuf.h"
#include "neuron.h"
#include "ns_plat.h"
#include "ni_mgmt2.h"

#include "XO2_dev.h"

#include "XO2_api.h"
#include "XO2_cmds.h"

#include "echelon.h"


#ifdef CONFIG_PROC_FS
  // enable or disable PROC_FS, if desired
  #define USE_PROC 01   // 0 or 1
#else
  #define USE_PROC 0    // ever 0
#endif


/* Version Information */
#define DRIVER_VERSION  "V3.23"
#define DRIVER_AUTHOR   "Volker Schober, vschober@gesytec.de"
#define DRIVER_DESC     "Easylon LVU Driver"
#define DRIVER_MOD_NAME "lvu"  // module name
#define DRIVER_DEV_NAME "lvu"  // /dev/lonusb
#define PROC_DIR_NAME	"easylon" // /proc/easylon
#define PROC_DEV_NAME	PROC_DIR_NAME "/" DRIVER_DEV_NAME  // /proc/easylon/lonusb


#define IOCTL_VERSION   0x43504C00


/* Get a minor range for your devices from the usb maintainer */
#define LONUSB_MINOR_BASE     64

/* we can have up to this number of device plugged in at once */
#define MAX_DEVICES             8
#define MAX_DEVICENODES         8
#define MAX_DEVICETABLE         (MAX_DEVICES * MAX_DEVICENODES)


#define USB_READBUFFER_SIZE     (256+64)    // 64*x
#define USB_WRITEBUFFER_SIZE    256         // 64*x



typedef enum _LPX_CARD
    {
    LPX_INVALID = -1,
    LVU001,
    GDS201,
    LVU2,
    } LPX_CARD;


#define DEBUG_DUMP1             0x0001
#define DEBUG_DUMP2             0x0002
#define DEBUG_DUMP_WATCHER      0x0004
#define DEBUG_OPEN_CLOSE        0x0008
#define DEBUG_OPEN_IOCTL        0x0010
#define DEBUG_VERBOSE           0x0080
#define DEBUG_REGISTER          0x0100
#define DEBUG_DISPATCH_READ     0x0200
#define DEBUG_DISPATCH_WRITE    0x0400
#define DEBUG_DUMP3             0x0800
#define DEBUG_BREAKPOINT        0x1000

#define FILE (strrchr(__FILE__, '/') ? strrchr(__FILE__, '/') + 1 : strrchr(__FILE__, '\\') ? strrchr(__FILE__, '\\') + 1 : __FILE__)
/* Use our own dbg macro */
#undef dbg
#define dbg(format, arg...) do { if (debug & DEBUG_VERBOSE) printk(KERN_DEBUG "%-13s: " format "\n" , FILE, ## arg); } while (0)
#define dbgz(zone, format, arg...) do { if (debug & (zone)) printk(KERN_DEBUG "%-13s: " format "\n" , FILE, ## arg); } while (0)
//#define dbg(format, arg...) do { if (debug) printk(KERN_INFO __FILE__ ": " format "\n" , ## arg); } while (0)
#ifndef err
  #define err(format, arg...) do { if (debug & DEBUG_REGISTER) printk(KERN_DEBUG "%-13s: " format "\n" , FILE, ## arg); } while (0)
#endif

#define Printf(_x_) {dbg _x_ ;}
#define DrvPrint(_x_) {dbg _x_ ;}
#define _DrvPrint(_x_) {dbg _x_ ;}


#define READ_IO_ULONG(ad)       ReadFpgaRegister (pdx->usbdev, ad)
#define WRITE_IO_ULONG(ad,val)  WriteFpgaRegister(pdx->usbdev, ad, val)


typedef struct _lonusb_dev_stats
{
	char	ProcDevName[sizeof(PROC_DEV_NAME)+10];

    __u64	open;
	__u64	close;

	__u64	usb_in;
	__u64	int_callback_ok;
	__u64	int_callback_abort;
	__u64	int_callback_error;

	__u64	submit_read1;
	__u64	submit_read1_ok;
	__u64	submit_read1_error;
	__u64	submit_read2;
	__u64	submit_read2_ok;
	__u64	submit_read2_error;

    __u64	read_local;
    __u64	read_in;
	__u64	read_ok;
	__u64	read_error;

    __u64	write_in;
    __u64	write_local;
	__u64	write_ok;
	__u64	write_error;

	__u64	poll;
	__u64	poll_read;
	__u64	poll_write;
	__u64	poll_error;
	__u64	poll_pollin;
	__u64	poll_pollout;
} lonusb_dev_stats;


/* Structure to hold all of our device specific stuff */
//typedef struct _SUB_DEVICE_EXTENSION SUB_DEVICE_EXTENSION, *PSUB_DEVICE_EXTENSION;

typedef struct _SUB_DEVICE_EXTENSION
{
    //PDEVICE_EXTENSION       udev;
    struct _DEVICE_EXTENSION *udev;

    unsigned char           SubdevNum;

    QUEUE                   ReadBufferQueue;        // Data Queue
    wait_queue_head_t       rd_wait;

    BYTE                    send_buffer[USB_WRITEBUFFER_SIZE];
    int32_t                 send_buffer_used;

    int                     open_count;

    lonusb_dev_stats        DevStats;
} SUB_DEVICE_EXTENSION, *PSUB_DEVICE_EXTENSION;


typedef struct _DEVICE_EXTENSION
{
    struct device           dev;                    // our device
    struct usb_device      *usbdev;                 // save off the usb device pointer
    struct usb_interface   *interface;              // the interface for this device
    unsigned char           minor;                  // the starting minor number for this device

    LPX_CARD                LpxCard;
    UCHAR                   RevisionID;

    unsigned char          *intr_in_buff;           // the buffer to receive data
    int                     bulk_in_size;           // the size of the receive buffer
    struct urb             *read_urb;               // the urb used to read data
    __u8                    intr_in_ep_addr;        // the address of the bulk in endpoint
    __u8                    bInterval;              // the polling interval of the bulk in endpoint

    wait_queue_head_t       wr_wait;
    UCHAR                   intdata[USB_READBUFFER_SIZE];
    __u32                   RemainDataSize;         // save incomplete Lon-Telegramm

    unsigned char          *bulk_out_buffer;        /* the buffer to send data */
    int32_t                 bulk_out_size;          /* the size of the send buffer */

    struct urb             *write_urb;              /* the urb used to send data */

    __u8                    bulk_out_endpointAddr;  /* the address of the bulk out endpoint */

    int32_t                 sub_open_count;
    struct mutex		    mutex; /* locks this structure */
    volatile int32_t        remove_pending;	        /* If set then about to exit */

    lonusb_dev_stats        DevStats;

    // *************************************************************
	Boolean_t		    cfgEn;    /**< 1=Configuration Logic access is enabled, 0=not */
	XO2Devices_t 	    devType;     /**< XO2 part number for information about sizes and programming times */
    DWORD               UserCode;
    // *************************************************************

    BYTE                neuron_id[NEURON_ID_LEN];
    domain_struct       Domain[2];

    BYTE                backlog;
    ULONG               PacketcycleTicks;
    ULONG               beta1PostRxTicks;
    ULONG               beta1PostTxTicks;
    ULONG               beta2Ticks;
    ULONG               beta2TicksFpga;
    ULONG               PreambleBits;       // counts only Bit Sync
    config_data_struct  Config;
    StatusStruct        Status;
    BYTE                error_log;          // last error

    BYTE                TxId;
    BYTE                reset_cause;        // use resetCause... macros below to eval
    BOOLEAN             HardwareWorking2;       // we should touch hardware at stop-device time
    u_int32_t           ServicePinTimeStamp;
    struct cdev         cdev;
    struct device      *SubDevices[MAX_DEVICENODES];

    SUB_DEVICE_EXTENSION  pSubDev[MAX_DEVICENODES];
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;


#define RAMSIZE     (128*1024)  // (128*1024)                  // RAM on LVU
#define RAMSIZE_GDS   (8*1024)  //  (10*1024)                  // RAM on GDS

// registers
  #define DATA_PORT                                    (0)
  #define LON_STAT_PORT                                (0x04)
  #define SET_RAM_ADDR_PORT                            (0x08)
  #define BYTE_COUNTER_PORT                            (0x0C)
  #define MSG_COUNTER_PORT                             (0x10)
  #define STAT_PORT                                    (0x14)
  #define CTRL_PORT                                    (0x14)
  #define PACKET_CYCLE_PORT                            (0x18)
  #define BAUD_AND_BIT_SYNC_PORT                       (0x20)
  #define RANDOM_READBACK_PORT                         (0x20)
  #define BETA_TIMER_PORT                              (0x1C)
  #define RAM_TEST_DATA_PORT                           (0x28)
  #define BETA2_TICKS_PORT                             (0x2C)
  #define CURR_RECV_TIME_PORT                          (0x2C)
  #define PREAMBLE_BITS_PORT                           (0x30)
  #define WRITE_RECV_TIME_PORT                         (0x34)
  #define DBACKLOG_PORT                                (0x34)
  #define BACKLOG_PORT                                 (0x38)
//  #define RECV_TIME_PORT                               (0x3C)


// register masks
//   LON
#define BETA1_PRIO_BIT              (1 << 27)

//     STAT_PORT
#define RCVD_MSG_IRQ                0x8000
#define RECV_TIME_READY             0x4000
#define SEND_MSG_IRQ                0x2000
#define KEY_IRQ                     0x0800
#define RA_HOST_WR                  0x0020

//     CTRL_PORT
#define HOSTWR_RDY                  0x0001
#define RCVD_RDY                    0x0002
#define HOSTWR_BYTE_TC_CLR          0x0004
#define NLON_RESET                  0x0008
#define RCVD_REL_IRQ                0x0010
#define LON_START_SEND              0x0020
#define HOST_RESET                  0x0040
#define BUMP_RCVD_MSG_CTR           0x0080
#define HOSTRD_RDY                  0x0100
#define HOST_PACKETCYCLE_RESTART    0x0200
#define SEND_REL_IRQ                0x0400
#define KEY_RES_STB                 0x0800
#define RCV_DIS_STB                 0x1000
#define FIFO_SET_FLU                0x2000
#define FIFO_RES_FLU                0x2000
#define DEBUG_STB                   0x8000

//     Backlog-Byte
#define CRC_ERROR_MASK              0x80
#define SERVICE_PIN_MASK            0x40
#define BACKLOG_MASK                0x3F



/* local function prototypes */
static ssize_t lvu_read        (struct file *file, char *buffer, size_t count, loff_t *ppos);
static ssize_t lvu_write       (struct file *file, const char *buffer, size_t count, loff_t *ppos);
static unsigned lvu_poll       (struct file *file, struct poll_table_struct *wait);
static long lvu_unlocked_ioctl (struct file *file, unsigned int cmd, unsigned long arg);
static int lvu_open            (struct inode *inode, struct file *file);
static int lvu_release         (struct inode *inode, struct file *file);

static int lvu_probe	    	(struct usb_interface *interface, const struct usb_device_id *id);
static void lvu_disconnect  	(struct usb_interface *interface);

static void lvu_read_int_callback      (struct urb *urb);
static void lvu_write_bulk_callback    (struct urb *urb);

static int CreateIntUrb(PDEVICE_EXTENSION dev);
static inline void lvu_delete (PDEVICE_EXTENSION dev);

static inline void lvu_debug_data (int DebugZone, const char *function, int size, const unsigned char *data);
static inline void dump_urb(struct urb * purb);

ULONG ReadFpgaRegister (struct usb_device *dev, ULONG RegIndex);
VOID WriteFpgaRegister(struct usb_device *dev, ULONG RegIndex, ULONG Value);

#endif
