/*++

Module Name:

    Neuron.cpp

Abstract:

    Neuron command emulation for Gesytec LVP

Environment:

    kernel mode only

Revision History:

    21.04.04 VS: Created

--*/

#include "xcvr.h"
#include "ni_comm.h"
#include "ni_mgmt.h"


//#warning brutal void pointer definition
//typedef void *PDEVICE_OBJECT;


#define COMM_PARAM_BYTES 11

#define EE_OFS_CURRENT_TXID         0xC0
#define EE_OFS_CURRENT_CONFIGDATA   0xC1
#define EE_OFS_TXID                 0xF0
#define EE_OFS_NID                  0xF2

////////////////////////////////////////////////////////////////////////
//
//  Data tables for transceiver parameter lookup

typedef struct {
    xcvr_ID     TxId;
    BYTE        Param_5[COMM_PARAM_BYTES];
    BYTE        Param_10[COMM_PARAM_BYTES];
    BYTE        Param_20[COMM_PARAM_BYTES];
} param_table;

static BOOLEAN reset_neuron(PDEVICE_EXTENSION pdx, BOOLEAN resmes, BOOLEAN doreset, BYTE reset_cause);
VOID InitNeuron(PDEVICE_EXTENSION pdx, BOOLEAN RebootNode);


                     //clks pin  prea pktc bet2 xmip rcip ndpr chpr fltr cdet
#define PARAM_NULL    {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}

#define NUM_PARAM_CHOICES 10
static const param_table params[NUM_PARAM_CHOICES] =
    {
        { TP_XF_78,     TP_XF_78_5,     TP_XF_78_10,    TP_XF_78_20   },
        { TP_XF_1250,   PARAM_NULL,     TP_XF_1250_10,  TP_XF_1250_20 },

        { FT_10,        FT_10_5,        FT_10_10,       FT_10_20      },

        { TP_RS485_39,  TP_RS_39_5,     TP_RS_39_10,    TP_RS_39_20   },
        { TP_RS485_78,  TP_RS_78_5,     TP_RS_78_10,    TP_RS_78_20   },
        { TP_RS485_625, TP_RS_625_5,    TP_RS_625_10,   TP_RS_625_20  },
        { TP_RS485_1250,PARAM_NULL,     TP_RS_1250_10,  TP_RS_1250_20 },

        { DC_78,        DC_78_5,        DC_78_10,       DC_78_20      },
        { DC_625,       DC_625_5,       DC_625_10,      DC_625_20     },
        { DC_1250,      PARAM_NULL,     DC_1250_10,     DC_1250_20    },

    };

BOOLEAN GetParamForTxId(config_data_struct *Config, BYTE TxId, Bits input_clock)
    {
      int i;

    for (i=0; i<NUM_PARAM_CHOICES; i++)
        {
        if (TxId == params[i].TxId)
            {
            const BYTE *Param_x;
            switch(input_clock)
                {
                case CLOCK_5000:  Param_x = params[i].Param_5;  break;
                case CLOCK_10000: Param_x = params[i].Param_10; break;
                case CLOCK_20000: Param_x = params[i].Param_20; break;
                default: return FALSE;
                }
            if (*Param_x)
                {
                  dbg ("using transceiver param_table %i with %i", i, input_clock);
                memcpy(Config->location + LOCATION_LEN, Param_x, COMM_PARAM_BYTES);
                return TRUE;
                }
            }
        }
    return FALSE;
    }

///////////////////////////////////////////////////////////////////////////////


VOID LoadNeuronData(PSUB_DEVICE_EXTENSION dev)
    // fdo == NULL: normal Call
    // fdo != NULL: Callback as WorkItem
    {/*
    NTSTATUS            ntStatus;
    // PLUGPLAY_REGKEY_DRIVER ->
    // W2K: HKLM\SYSTEM\CCS\Class\{*}\0006\*
    //      --> Ndi\params\*
    // W98: HKLM\SYSTEM\CCS\Services\Class\USB\0002\*
    //      --> Ndi\params\*

    if (KeGetCurrentIrql() == PASSIVE_LEVEL)
        {
        if (fdo)
            {
            KIRQL oldirql;
            KeAcquireSpinLock(&dev->polllock, &oldirql);
            if (dev->LoadNeuronDataItem)
                {
                IoFreeWorkItem(dev->LoadNeuronDataItem);
                dev->LoadNeuronDataItem = NULL;
                }
            KeReleaseSpinLock(&dev->polllock, oldirql);
            }
        }
    else
        {
        dbg ("LoadNeuronData() later: IRQL != PASSIVE_LEVEL");
        KIRQL oldirql;
        KeAcquireSpinLock(&dev->polllock, &oldirql);
        if (!dev->LoadNeuronDataItem)
            {
            dev->LoadNeuronDataItem = IoAllocateWorkItem(dev->DeviceObject);
            KeReleaseSpinLock(&dev->polllock, oldirql);

            if (dev->LoadNeuronDataItem)
                {
                // Queue the work item for delayed execution. IoQueueWorkItem will take out an
                // extra reference to our device object that will persist until after the
                // callback routine returns, thereby preventing this driver from unloading before
                // the last instruction in this driver finishes executing.

                IoQueueWorkItem(dev->LoadNeuronDataItem, (PIO_WORKITEM_ROUTINE) LoadNeuronData, DelayedWorkQueue, dev);
                }
            else
                {
                dbg("AllocateWorkItem failed");
                }
            }
        else
            KeReleaseSpinLock(&dev->polllock, oldirql);
        }*/
    }


VOID SaveNeuronData(PDEVICE_EXTENSION pdx)
    // fdo == NULL: normal Call
    // fdo != NULL: Callback as WorkItem
    {
    dbg("SaveNeuronData()");
//
#ifndef __KERNEL__
    ULONG PortStatus = GetStatus(pdx);

    if (PortStatus == 0)  return;
#endif

//
    EepromAccess(pdx->usbdev, &(pdx->TxId),   EE_OFS_CURRENT_TXID,       sizeof(pdx->TxId),   FALSE);  // Write
    EepromAccess(pdx->usbdev, &(pdx->Config), EE_OFS_CURRENT_CONFIGDATA, sizeof(pdx->Config), FALSE);  // Write
    }


///////////////////////////////////////////////////////////////////////////////


BYTE HandleQueryStatus(PDEVICE_EXTENSION pdx, ExplicitMsg *Request, ExplicitMsg *Response)
    {  // ND_query_status
    status_struct *Status = (status_struct *)Response->data;
    memset(Status, 0, sizeof(*Status));
    memcpy(Status, &pdx->Status, sizeof(base_stats));
    Status->reset_cause     = pdx->reset_cause;

    if (pdx->LpxCard == GDS201 /*|| pdx->LpxCard == GDS203*/ || pdx->LpxCard == LVU2 /*|| pdx->LpxCard == LVX4*/)
        {
#if 0
        int status;
        DWORD         UserCode;   // 32 bit user code set in XO2
        unsigned int  Stat;

        XO2ECA_apiJEDECinfo(pdx, &XO2_JEDEC_GDS);

        status = XO2ECA_apiGetDevType(pdx);
  #if 01
        status = XO2ECAcmd_Bypass(pdx);

        status = XO2ECAcmd_readUserCode(pdx, &UserCode);
        _DrvPrint (("XO2ECAcmd_readUserCode()=%d UserCode=%08X\n", status, UserCode));

        status = XO2ECAcmd_readStatusReg(pdx, &Stat);
        _DrvPrint (("XO2ECAcmd_readStatusReg()=%d Stat=%08X\n", status, Stat));

        {
        XO2RegInfo_t Info;
        status = XO2ECA_apiGetHdwInfo(pdx, &Info);
        _DrvPrint (("XO2ECA_apiGetHdwInfo()=%d devID=%08X UserCode=%08X TraceID=%02X%02X%02X%02X%02X%02X%02X%02X\n", 
            status, Info.devID, Info.UserCode, 
            Info.TraceID[0], Info.TraceID[1], Info.TraceID[2], Info.TraceID[3], 
            Info.TraceID[4], Info.TraceID[5], Info.TraceID[6], Info.TraceID[7]
            ));
        

        status = XO2ECAcmd_openCfgIF(pdx, TRANSPARENT_MODE);
        
        status = XO2ECAcmd_readStatusReg(pdx, &Stat);
        _DrvPrint (("XO2ECAcmd_readStatusReg()=%d Stat=%08X\n", status, Stat));

        status = XO2ECAcmd_readUserCode(pdx, &UserCode);
        _DrvPrint (("XO2ECAcmd_readUserCode()=%d UserCode=%08X\n", status, UserCode));
        
        status = XO2ECAcmd_closeCfgIF(pdx);
        status = XO2ECAcmd_Bypass(pdx);


        status = XO2ECA_apiGetHdwInfo(pdx, &Info);
        _DrvPrint (("XO2ECA_apiGetHdwInfo()=%d devID=%08X UserCode=%08X TraceID=%02X%02X%02X%02X%02X%02X%02X%02X\n", 
            status, Info.devID, Info.UserCode, 
            Info.TraceID[0], Info.TraceID[1], Info.TraceID[2], Info.TraceID[3], 
            Info.TraceID[4], Info.TraceID[5], Info.TraceID[6], Info.TraceID[7]
            ));
        }
  #endif
#endif
        }

#if defined DBG
        {
        VENDOR_REQUEST_IN VendorRequest =
            //Req   Value   Index   Len                                    1=in
            { VR_FW_VERSION, 0x0000, 0x0000, sizeof(VendorRequest.buffer), 1};
        ULONG ret = Ezusb_VendorRequest(pdx->DeviceObject, &VendorRequest);

        dbg ("Ezusb_VendorRequest(VR_FW_VERSION)=%X buffer=%02X %02X %02X %02X %02X %02X %02X %02X", ret,
            VendorRequest.buffer[0], VendorRequest.buffer[1],
            VendorRequest.buffer[2], VendorRequest.buffer[3],
            VendorRequest.buffer[4], VendorRequest.buffer[5],
            VendorRequest.buffer[6], VendorRequest.buffer[7]);
        }
#endif
#if 1 //defined DBG
        {
            ULONG TimeStamp     = READ_IO_ULONG(CURR_RECV_TIME_PORT);
            ULONG Stat          = READ_IO_ULONG(STAT_PORT);
            ULONG LonStat       = READ_IO_ULONG(LON_STAT_PORT);
            ULONG ByteCounter   = READ_IO_ULONG(BYTE_COUNTER_PORT);
            ULONG MsgCounter    = READ_IO_ULONG(MSG_COUNTER_PORT);
            ULONG Backlog       = READ_IO_ULONG(BACKLOG_PORT);
            
            err("QueryStatus(1): TimeStamp=%08X LonStat=%08X Stat=%08X ByteCounter=%02X  MsgCounter=%02X Backlog=%02X",
                TimeStamp, LonStat, Stat, ByteCounter, MsgCounter, Backlog);

//     TimeStamp     = READ_IO_ULONG(CURR_RECV_TIME_PORT);
//     Stat          = READ_IO_ULONG(STAT_PORT);
//     LonStat       = READ_IO_ULONG(LON_STAT_PORT);
//     ByteCounter   = READ_IO_ULONG(BYTE_COUNTER_PORT);
//     MsgCounter    = READ_IO_ULONG(MSG_COUNTER_PORT);
//     Backlog       = READ_IO_ULONG(BACKLOG_PORT);
//     dbg ("QueryStatus(2): TimeStamp=%08X LonStat=%08X Stat=%08X ByteCounter=%02X  MsgCounter=%02X Backlog=%02X",
//         TimeStamp, LonStat, Stat, ByteCounter, MsgCounter, Backlog);
        }
#endif

    Status->node_state  = CNFG_ONLINE;
    if (pdx->TxId == 0xFF)
        {
        Status->node_state  = APPL_UNCNFG;
        Status->error_log   = INVALID_XCVR_REG_ADDR;    // Invalid transceiver register address
        }
    else if (0 == memcmp(pdx->neuron_id, "\xFF\xFF\xFF\xFF\xFF\xFF", NEURON_ID_LEN))
        {
        Status->node_state  = APPL_UNCNFG;
        Status->error_log   = BAD_ADDRESS_TYPE;         // Bad address type
        }
    if (pdx->error_log)
        Status->error_log   = pdx->error_log;
    if (!pdx->HardwareWorking2)
        {
        Status->node_state  = Status->node_state == CNFG_ONLINE ? CNFG_OFFLINE : BLANK;
//        Status->error_log   = SELF_TEST_FAILED;       // Self test failed
//        Status->error_log   = MEMORY_ALLOC_FAILURE;   // RAM memory allocation failure
        }
    Status->version_number  = 124;  // VNI
    Status->model_number    = MODEL_3150;

    Response->code = ND_query_status_succ;
    return sizeof(status_struct) +1;
    }


BYTE HandleClearStatus(PDEVICE_EXTENSION pdx, ExplicitMsg *Request, ExplicitMsg *Response)
    {  // ND_clear_status
    memset(&pdx->Status, 0, sizeof(pdx->Status));
    pdx->reset_cause = CLEARED;     // POWER_UP_RESET
    pdx->error_log   = NO_ERRORS;   // 0

    Response->code = ND_clear_status_succ;
    return 1;
    }


BYTE HandleReadMemory(PDEVICE_EXTENSION pdx, ExplicitMsg *Request, ExplicitMsg *Response, WORD ChildIndex)
    {  // NM_read_memory
    NM_read_memory_request *MemRequest = (NM_read_memory_request*)Request;
    Word Offset = NET_SWAB(MemRequest->offset);
    Byte Count  = MemRequest->count;
    static const Byte id_string[ID_STR_LEN+1] = "EASYLON2";
    read_only_data_struct ReadOnly;

    if (Count> sizeof(Response->data))
        return 1;

    memset(&ReadOnly, 0, sizeof(ReadOnly));
    memcpy(ReadOnly.neuron_id, pdx->neuron_id, NEURON_ID_LEN);
    ReadOnly.neuron_id[NEURON_ID_LEN-1] += ChildIndex;
    memcpy(ReadOnly.id_string, id_string, ID_STR_LEN);
    ReadOnly.model_num          = MODEL_3150;
    ReadOnly.minor_model_num    = pdx->RevisionID;
    ReadOnly.two_domains        = 1;
    ReadOnly.receive_trans_count = 7;
    ReadOnly.app_buf_in_count = ReadOnly.app_buf_out_count = ReadOnly.app_buf_out_priority_count = 3;
    ReadOnly.net_buf_in_count = 12;
    ReadOnly.net_buf_out_count = ReadOnly.net_buf_out_priority_count = 8;

    memset(Response->data, 0, Count);

    switch (MemRequest->mode)
        {
        case ABSOLUTE_MEM_MODE:
            dbg ("ABSOLUTE_MEM_MODE: Offset=%04X Count=%02X", Offset, Count);
            if (Offset >= OFS_STAT_3150_V7)
                {
                if (Offset < ABS_READ_ONLY)
                    {
                    Offset -= OFS_STAT_3150_V7;
                    goto StatsRelative;
                    }
                if (Offset < ABS_CONFIG)
                    {
                    Offset -= ABS_READ_ONLY;
                    goto ReadOnlyRelative;
                    }
                if (Offset < ABS_CONFIG+sizeof(config_data_struct))
                    {
                    Offset -= ABS_CONFIG;
                    goto ConfigRelative;
                    }
                }
            if (Offset == OFS_FW_VERSION)  // 0
                {
                if (Count == 1)
                    Response->data[0] = 11;  // Base Firmware
                else
                    Response->data[0] = 124; // Firmware VNI
                }
            else if (Offset == BOOT_ID_ADDR)
                {
                Response->data[0] = HIBYTE(pdx->UserCode);
                Response->data[1] = LOBYTE(pdx->UserCode);
                }
            break;

        case STATS_RELATIVE:      // = absolute 0xE904  size 0x0A+0x10 = 0x1A
StatsRelative:
            dbg ("STATS_RELATIVE: Offset=%04X Count=%02X  sizeof(status_struct)=%04X sizeof(network_stats)=%04X",
                Offset, Count, (unsigned int)sizeof(status_struct), (unsigned int)sizeof(network_stats));
            if (Offset < sizeof(pdx->Status))
                memcpy(Response->data, (PUCHAR)(&(pdx->Status)) + Offset, MIN(Count, sizeof(pdx->Status) - Offset));
            break;
        case READ_ONLY_RELATIVE:  // = absolute 0xF000  size 0x24+0x05 = 0x29
ReadOnlyRelative:
            dbg ("READ_ONLY_RELATIVE: Offset=%04X Count=%02X  sizeof(read_only_data_struct)=%04X sizeof(read_only_data_struct_2)=%04X",
                Offset, Count, (unsigned int)sizeof(read_only_data_struct), (unsigned int)sizeof(read_only_data_struct_2));
            if (Offset < sizeof(ReadOnly))
                memcpy(Response->data, (PUCHAR)(&(ReadOnly))+Offset, MIN(Count, sizeof(ReadOnly)-Offset));
            break;
        case CONFIG_RELATIVE:     // = absolute 0xF029  size 0x19
ConfigRelative:
            dbg ("CONFIG_RELATIVE: Offset=%04X Count=%02X  sizeof(config_data_struct)=%04X",
                Offset, Count, (unsigned int)sizeof(config_data_struct));
            if (Offset < sizeof(pdx->Config))
                memcpy(Response->data, (PUCHAR)(&(pdx->Config))+Offset, MIN(Count, sizeof(pdx->Config)-Offset));
            break;

        default:
            return 1;
        }

    Response->code = NM_read_memory_success;
    return Count +1;
    }


BYTE HandleWriteMemory(PDEVICE_EXTENSION pdx, ExplicitMsg *Request, ExplicitMsg *Response)
    {  // NM_write_memory
    NM_write_memory_request_nc *MemRequest = (NM_write_memory_request_nc*)Request;
    Word    Offset  = NET_SWAB(MemRequest->offset);
    Byte    Count   = MemRequest->count;
    Byte   *data    = (Byte*)(MemRequest +1);
    if (Count> sizeof(Response->data))
        return 1;

    switch (MemRequest->mode)
        {
        case ABSOLUTE_MEM_MODE:
            dbg ("ABSOLUTE_MEM_MODE: Offset=%04X Count=%02X", Offset, Count);
            if (Offset >= OFS_STAT_3150_V7)
                {
                if (Offset < ABS_READ_ONLY)
                    {
                    Offset -= OFS_STAT_3150_V7;
                    goto StatsRelative;
                    }
                if (Offset < ABS_CONFIG)
                    {
                    Offset -= ABS_READ_ONLY;
                    goto ReadOnlyRelative;
                    }
                if (Offset < ABS_CONFIG+sizeof(config_data_struct))
                    {
                    Offset -= ABS_CONFIG;
                    goto ConfigRelative;
                    }
                }
            break;

        case STATS_RELATIVE:      // = absolute 0xE904  size 0x0A+0x10 = 0x1A
StatsRelative:
            dbg ("sizeof(status_struct)=%04X sizeof(network_stats)=%04X", (unsigned int)sizeof(status_struct), (unsigned int)sizeof(network_stats));
            break;
        case READ_ONLY_RELATIVE:  // = absolute 0xF000  size 0x24+0x05 = 0x29
ReadOnlyRelative:
            dbg ("sizeof(read_only_data_struct)=%04X sizeof(read_only_data_struct_2)=%04X", (unsigned int)sizeof(read_only_data_struct), (unsigned int)sizeof(read_only_data_struct_2));
            break;
        case CONFIG_RELATIVE:     // = absolute 0xF029  size 0x19
ConfigRelative:
            dbg ("CONFIG_RELATIVE: Offset=%04X Count=%02X  sizeof(config_data_struct)=%04X",
                Offset, Count, (unsigned int)sizeof(config_data_struct));
            if (Offset == OFS_COMM_TYPE && Count && data[0] == 0)
                {  // Reboot node
                InitNeuron(pdx, TRUE);
                }
            else
                {
                if (Offset < sizeof(pdx->Config))
                    memcpy((PUCHAR)(&(pdx->Config))+Offset, data, MIN (Count, sizeof(pdx->Config)-Offset));
                CalculateTimingParameters(pdx);
                }
            SaveNeuronData(pdx);
            break;

        default:
            return 1;
        }

    Response->code = NM_write_memory_success;
    return 1;
    }


BYTE HandleDeviceEsc(PDEVICE_EXTENSION pdx, ExplicitMsg *Request, ExplicitMsg *Response)
    {  // NM_device_esc
    NM_device_request *DevRequest = (NM_device_request*)Request;
    if (DevRequest->devReqHdr.device_code == PRODUCT_QUERY_PRODUCT_CODE &&
        DevRequest->devReqHdr.device_command == PRODUCT_QUERY_COMMAND)
        {
        NM_product_query_response *ProdResponse = (NM_product_query_response*)Response;
        ProdResponse->product           = 2;
        ProdResponse->model             = 3;
        ProdResponse->version           = 1;
        ProdResponse->product_specific  = 0;
        //ProdResponse->XID_reserved      = 0;

        ProdResponse->transceiver_ID    = pdx->TxId;
        }
    else
        return 1;
    Response->code = NM_device_esc_success;
    return sizeof(NM_product_query_response);
    }

BYTE HandleQueryDomain(PDEVICE_EXTENSION pdx, ExplicitMsg *Request, ExplicitMsg *Response)
    {  // NM_query_domain
    NM_query_domain_request *DevRequest = (NM_query_domain_request*)Request;
    if (DevRequest->domain_index <= 1)
        {
        NM_query_domain_response *ProdResponse = (NM_query_domain_response*)Response;
        memcpy(&ProdResponse->id, &pdx->Domain[DevRequest->domain_index], sizeof(domain_struct));
        }
    else
        return 1;
    Response->code = NM_query_domain_success;
    return sizeof(NM_query_domain_response);
    }

BYTE HandleUpdateDomain(PDEVICE_EXTENSION pdx, ExplicitMsg *Request, ExplicitMsg *Response)
    {  // NM_update_domain
    NM_update_domain_request *DevRequest = (NM_update_domain_request*)Request;
    if (DevRequest->domain_index <= 1)
        {
        memcpy(&pdx->Domain[DevRequest->domain_index], &DevRequest->id, sizeof(domain_struct));
        }
    else
        return 1;
    Response->code = NM_update_domain_success;
    return 1;
    }


///////////////////////////////////////////////////////////////////////////////


VOID InitNeuron(PDEVICE_EXTENSION pdx, BOOLEAN RebootNode)
    {
    dbg("InitNeuron():");

    EepromAccess(pdx->usbdev, &(pdx->TxId),     EE_OFS_TXID, sizeof(pdx->TxId),     TRUE);   // Read

    dbg ("txid : %02X", ((unsigned int)(pdx->TxId)));

    EepromAccess(pdx->usbdev, pdx->neuron_id, EE_OFS_NID, sizeof(pdx->neuron_id), TRUE);   // Read

    dbg ("neuron id: %02X:%02X:%02X:%02X:%02X:%02X",
         ((unsigned int)(pdx->neuron_id[0])), ((unsigned int)(pdx->neuron_id[1])),
         ((unsigned int)(pdx->neuron_id[2])), ((unsigned int)(pdx->neuron_id[3])),
         ((unsigned int)(pdx->neuron_id[4])), ((unsigned int)(pdx->neuron_id[5])));

    GetParamForTxId(&(pdx->Config), pdx->TxId, CLOCK_10000);

    if (pdx->TxId != 0xFF)
        {
        BYTE CurrentTxId;
        EepromAccess(pdx->usbdev, &CurrentTxId, EE_OFS_CURRENT_TXID, sizeof(CurrentTxId), TRUE);  // Read
        if (CurrentTxId == 0xFF || RebootNode)
            {
            SaveNeuronData(pdx);
            }
        else
            {
            pdx->TxId = CurrentTxId;
            GetParamForTxId(&(pdx->Config), pdx->TxId, CLOCK_10000);
            EepromAccess(pdx->usbdev, &(pdx->Config), EE_OFS_CURRENT_CONFIGDATA, sizeof(pdx->Config), TRUE);  // Read

            // LoadNeuronData(NULL, pdx);
            }
        }

    CalculateTimingParameters(pdx);

    pdx->Domain[0].id[0]        = 0x08;
    pdx->Domain[0].subnet       = 1;
    pdx->Domain[0].node         = 1;
    pdx->Domain[0].cloneDomain  = 1;
    pdx->Domain[0].len          = 1;
    memset(&(pdx->Domain[0].key), 0xFF, AUTH_KEY_LEN);
    pdx->Domain[1].cloneDomain  = 1;
    pdx->Domain[1].len          = NULL_DOMAIN_LEN;
    memset(&(pdx->Domain[1].key), 0xFF, AUTH_KEY_LEN);
    }


VOID ExitNeuron(PDEVICE_EXTENSION pdx)
    {
    dbg ("ExitNeuron():");
//    if (KeGetCurrentIrql() != PASSIVE_LEVEL)  return;
    if (pdx->HardwareWorking2)
        SaveNeuronData(pdx);
    }


ULONG Neuron(PDEVICE_EXTENSION pdx, ExpAppBuffer *InMsg, ULONG InLen, ExpAppBuffer * OutMsg, WORD ChildIndex)
    {
    //dbg ("Neuron(): InMsg->ni_hdr.noq.cmd=%02X InLen=%02X", InMsg->ni_hdr.noq.cmd, InLen);
    ULONG OutLen = 0;
    if ((InMsg->ni_hdr.q.queue == niTQ || InMsg->ni_hdr.q.queue == niTQ_P || InMsg->ni_hdr.q.queue == niNTQ || InMsg->ni_hdr.q.queue == niNTQ_P)
        && InMsg->ni_hdr.q.q_cmd == niNETMGMT)
        {
        BYTE ResponeLen = 1;
        memcpy(OutMsg, InMsg, sizeof(ExpAppBuffer)-sizeof(MsgData));
        OutMsg->data.exp.code = InMsg->data.exp.code & ~0x60;
        dbg ("lon message code: %02X", InMsg->data.exp.code);
        switch(InMsg->data.exp.code)
            {
            case ND_query_status:
                ResponeLen = HandleQueryStatus(pdx, &InMsg->data.exp, &OutMsg->data.exp);
                goto CompleteMsg;
            case ND_clear_status:
                ResponeLen = HandleClearStatus(pdx, &InMsg->data.exp, &OutMsg->data.exp);
                goto CompleteMsg;

            case NM_read_memory:
                ResponeLen = HandleReadMemory(pdx, &InMsg->data.exp, &OutMsg->data.exp, ChildIndex);
                goto CompleteMsg;
            case NM_write_memory:
                ResponeLen = HandleWriteMemory(pdx, &InMsg->data.exp, &OutMsg->data.exp);
                goto CompleteMsg;

            case NM_query_domain:
                ResponeLen = HandleQueryDomain(pdx, &InMsg->data.exp, &OutMsg->data.exp);
                goto CompleteMsg;
            case NM_update_domain:
                ResponeLen = HandleUpdateDomain(pdx, &InMsg->data.exp, &OutMsg->data.exp);
                goto CompleteMsg;

            case NM_device_esc:
                ResponeLen = HandleDeviceEsc(pdx, &InMsg->data.exp, &OutMsg->data.exp);

            default:
CompleteMsg:
				//dbg ("CompleteMsg: st=%X=%d REQUEST=%X=%d", InMsg->msg_hdr.exp.st, InMsg->msg_hdr.exp.st, REQUEST, REQUEST);
                if (InMsg->msg_hdr.exp.st != REQUEST)  return 0;  // only answer for request
                OutMsg->msg_hdr.exp.length = ResponeLen;
                OutMsg->msg_hdr.exp.response  = 1;
                OutMsg->ni_hdr.q.queue    = niRESPONSE;
                OutMsg->ni_hdr.q.q_cmd    = niCOMM;
                //OutMsg->ni_hdr.noq.cmd    = niDRV_CMD;
#if defined ADDRESS_EXTENSION
                OutMsg->ni_hdr.q.length   = OutMsg->msg_hdr.exp.length + 0x12;
#else
                OutMsg->ni_hdr.q.length   = OutMsg->msg_hdr.exp.length + 0x0E;
#endif
                OutLen = OutMsg->ni_hdr.q.length+2;
                break;
            }
        }
    else
        {
        OutLen = MIN(InLen, sizeof(ExpAppBuffer));
        memcpy(OutMsg, InMsg, OutLen);
        //dbg ("Neuron(): InLen=%02X OutLen=%02X sizeof(ExpAppBuffer)=%02X", InLen, OutLen, sizeof(ExpAppBuffer));

        switch (InMsg->ni_hdr.noq.cmd)
            {
            case niFLUSH_CANCEL:
            case niSERVICE:
                return 0;       // ignore
            case niRESET:
                if (reset_neuron(pdx, TRUE, TRUE, SOFTWARE_RESET))
                    {
                    pdx->HardwareWorking2 = TRUE;
                    //EnableIrq(pdx, TRUE);
                    }
                else
                    {
                    //EnableIrq(pdx, FALSE);
                    pdx->error_log   = APPL_SIGNATURE_BAD;  // Application image inconsistency
                    pdx->HardwareWorking2 = FALSE;
                    }
                break;
            case niTXID:
                if (InMsg->ni_hdr.noq.length == 1)
                    {
                    if (InMsg->msg_hdr.noqData[0] == XCVR_UNKNOWN)
                        InMsg->msg_hdr.noqData[0] = pdx->TxId;
                    else
                        {
                        pdx->TxId = InMsg->msg_hdr.noqData[0];
                        GetParamForTxId(&pdx->Config, pdx->TxId, CLOCK_10000);
                        CalculateTimingParameters(pdx);
                        SaveNeuronData(pdx);
                        }
                    }
                break;
            }
        }
    //dbg ("Neuron(): return %02X", OutLen);
    return OutLen;
    }							// Neuron

/*******************************************************************************
Function:  CRC16
Returns:   16 bit CRC computed.
Purpose:   To compute the 16 bit CRC for a given buffer.
Comments:  None.
*******************************************************************************/

/* CRC16 implementation acording to CCITT standards
 *
 * Polynom: x16 + x12 + x5 + 1
 */

static const USHORT ccitt_crc16_lookup[256]= {
  0x0000,0x1021,0x2042,0x3063,0x4084,0x50a5,0x60c6,0x70e7,
  0x8108,0x9129,0xa14a,0xb16b,0xc18c,0xd1ad,0xe1ce,0xf1ef,
  0x1231,0x0210,0x3273,0x2252,0x52b5,0x4294,0x72f7,0x62d6,
  0x9339,0x8318,0xb37b,0xa35a,0xd3bd,0xc39c,0xf3ff,0xe3de,
  0x2462,0x3443,0x0420,0x1401,0x64e6,0x74c7,0x44a4,0x5485,
  0xa56a,0xb54b,0x8528,0x9509,0xe5ee,0xf5cf,0xc5ac,0xd58d,
  0x3653,0x2672,0x1611,0x0630,0x76d7,0x66f6,0x5695,0x46b4,
  0xb75b,0xa77a,0x9719,0x8738,0xf7df,0xe7fe,0xd79d,0xc7bc,
  0x48c4,0x58e5,0x6886,0x78a7,0x0840,0x1861,0x2802,0x3823,
  0xc9cc,0xd9ed,0xe98e,0xf9af,0x8948,0x9969,0xa90a,0xb92b,
  0x5af5,0x4ad4,0x7ab7,0x6a96,0x1a71,0x0a50,0x3a33,0x2a12,
  0xdbfd,0xcbdc,0xfbbf,0xeb9e,0x9b79,0x8b58,0xbb3b,0xab1a,
  0x6ca6,0x7c87,0x4ce4,0x5cc5,0x2c22,0x3c03,0x0c60,0x1c41,
  0xedae,0xfd8f,0xcdec,0xddcd,0xad2a,0xbd0b,0x8d68,0x9d49,
  0x7e97,0x6eb6,0x5ed5,0x4ef4,0x3e13,0x2e32,0x1e51,0x0e70,
  0xff9f,0xefbe,0xdfdd,0xcffc,0xbf1b,0xaf3a,0x9f59,0x8f78,
  0x9188,0x81a9,0xb1ca,0xa1eb,0xd10c,0xc12d,0xf14e,0xe16f,
  0x1080,0x00a1,0x30c2,0x20e3,0x5004,0x4025,0x7046,0x6067,
  0x83b9,0x9398,0xa3fb,0xb3da,0xc33d,0xd31c,0xe37f,0xf35e,
  0x02b1,0x1290,0x22f3,0x32d2,0x4235,0x5214,0x6277,0x7256,
  0xb5ea,0xa5cb,0x95a8,0x8589,0xf56e,0xe54f,0xd52c,0xc50d,
  0x34e2,0x24c3,0x14a0,0x0481,0x7466,0x6447,0x5424,0x4405,
  0xa7db,0xb7fa,0x8799,0x97b8,0xe75f,0xf77e,0xc71d,0xd73c,
  0x26d3,0x36f2,0x0691,0x16b0,0x6657,0x7676,0x4615,0x5634,
  0xd94c,0xc96d,0xf90e,0xe92f,0x99c8,0x89e9,0xb98a,0xa9ab,
  0x5844,0x4865,0x7806,0x6827,0x18c0,0x08e1,0x3882,0x28a3,
  0xcb7d,0xdb5c,0xeb3f,0xfb1e,0x8bf9,0x9bd8,0xabbb,0xbb9a,
  0x4a75,0x5a54,0x6a37,0x7a16,0x0af1,0x1ad0,0x2ab3,0x3a92,
  0xfd2e,0xed0f,0xdd6c,0xcd4d,0xbdaa,0xad8b,0x9de8,0x8dc9,
  0x7c26,0x6c07,0x5c64,0x4c45,0x3ca2,0x2c83,0x1ce0,0x0cc1,
  0xef1f,0xff3e,0xcf5d,0xdf7c,0xaf9b,0xbfba,0x8fd9,0x9ff8,
  0x6e17,0x7e36,0x4e55,0x5e74,0x2e93,0x3eb2,0x0ed1,0x1ef0
};

USHORT CRC16(PBYTE bufInOut, USHORT len)
    {
    USHORT crc = 0xFFFF;
    USHORT i;

    for (i = 0; i < len; i++)
        crc = (crc<<8) ^ ccitt_crc16_lookup[((crc>>8) ^ bufInOut[i]) & 0x00FF];

    crc ^= 0xFFFF;
    return crc;
    }
